from pathlib import Path


def _create_dir(dir_name: str) -> None:
    """
    :param dir_name: create dir use dir name
    :return: None
    """
    Path(dir_name).mkdir(
        parents=True,
        exist_ok=True
    )


def create_template_dir() -> None:
    _create_dir("web_runner/template")



import sys
import types
from typing import Tuple

from je_web_runner.je_web_runner.webrunner_manager import web_runner
from je_web_runner.utils.exception.exception_tag import add_command_exception_tag
from je_web_runner.utils.exception.exception_tag import executor_data_error, executor_list_error
from je_web_runner.utils.exception.exceptions import WebRunnerExecuteException, WebRunnerAddCommandException
from je_web_runner.utils.html_report.html_report_generate import generate_html
from je_web_runner.utils.json.json_file.json_file import read_action_json
from je_web_runner.utils.test_object.test_object_record.test_object_record_class import test_object_record
from je_web_runner.utils.test_record.test_record_class import test_record_instance


class Executor(object):

    def __init__(self):
        self.event_dict = {
            # webdriver manager
            "get_webdriver_manager": web_runner.new_driver,
            "change_index_of_webdriver": web_runner.change_webdriver,
            "quit": web_runner.quit,
            # test object
            "SaveTestObject": test_object_record.save_test_object,
            "CleanTestObject": test_object_record.clean_record,
            # webdriver wrapper
            "set_driver": web_runner.webdriver_wrapper.set_driver,
            "set_webdriver_options_capability": web_runner.webdriver_wrapper.set_driver,
            "find_element": web_runner.webdriver_wrapper.find_element_with_test_object_record,
            "find_elements": web_runner.webdriver_wrapper.find_elements_with_test_object_record,
            "implicitly_wait": web_runner.webdriver_wrapper.implicitly_wait,
            "explict_wait": web_runner.webdriver_wrapper.explict_wait,
            "to_url": web_runner.webdriver_wrapper.to_url,
            "forward": web_runner.webdriver_wrapper.forward,
            "back": web_runner.webdriver_wrapper.back,
            "refresh": web_runner.webdriver_wrapper.refresh,
            "switch": web_runner.webdriver_wrapper.switch,
            "set_script_timeout": web_runner.webdriver_wrapper.set_script_timeout,
            "set_page_load_timeout": web_runner.webdriver_wrapper.set_page_load_timeout,
            "get_cookies": web_runner.webdriver_wrapper.get_cookies,
            "get_cookie": web_runner.webdriver_wrapper.get_cookie,
            "add_cookie": web_runner.webdriver_wrapper.add_cookie,
            "delete_cookie": web_runner.webdriver_wrapper.delete_cookie,
            "delete_all_cookies": web_runner.webdriver_wrapper.delete_all_cookies,
            "execute": web_runner.webdriver_wrapper.execute,
            "execute_script": web_runner.webdriver_wrapper.execute_script,
            "execute_async_script": web_runner.webdriver_wrapper.execute_async_script,
            "move_to_element": web_runner.webdriver_wrapper.move_to_element_with_test_object,
            "move_to_element_with_offset": web_runner.webdriver_wrapper.move_to_element_with_offset_and_test_object,
            "drag_and_drop": web_runner.webdriver_wrapper.drag_and_drop_with_test_object,
            "drag_and_drop_offset": web_runner.webdriver_wrapper.drag_and_drop_offset_with_test_object,
            "perform": web_runner.webdriver_wrapper.perform,
            "reset_actions": web_runner.webdriver_wrapper.reset_actions,
            "left_click": web_runner.webdriver_wrapper.left_click_with_test_object,
            "left_click_and_hold": web_runner.webdriver_wrapper.left_click_and_hold_with_test_object,
            "right_click": web_runner.webdriver_wrapper.right_click_with_test_object,
            "left_double_click": web_runner.webdriver_wrapper.left_double_click_with_test_object,
            "release": web_runner.webdriver_wrapper.release_with_test_object,
            "press_key": web_runner.webdriver_wrapper.press_key_with_test_object,
            "release_key": web_runner.webdriver_wrapper.release_key_with_test_object,
            "move_by_offset": web_runner.webdriver_wrapper.move_by_offset,
            "pause": web_runner.webdriver_wrapper.pause,
            "send_keys": web_runner.webdriver_wrapper.send_keys,
            "send_keys_to_element": web_runner.webdriver_wrapper.send_keys_to_element_with_test_object,
            "scroll": web_runner.webdriver_wrapper.scroll,
            "check_current_webdriver": web_runner.webdriver_wrapper.check_current_webdriver,
            "maximize_window": web_runner.webdriver_wrapper.maximize_window,
            "fullscreen_window": web_runner.webdriver_wrapper.fullscreen_window,
            "minimize_window": web_runner.webdriver_wrapper.minimize_window,
            "set_window_size": web_runner.webdriver_wrapper.set_window_size,
            "set_window_position": web_runner.webdriver_wrapper.set_window_position,
            "get_window_position": web_runner.webdriver_wrapper.get_window_position,
            "get_window_rect": web_runner.webdriver_wrapper.get_window_rect,
            "set_window_rect": web_runner.webdriver_wrapper.set_window_rect,
            "get_screenshot_as_png": web_runner.webdriver_wrapper.get_screenshot_as_png,
            "get_screenshot_as_base64": web_runner.webdriver_wrapper.get_screenshot_as_base64,
            "get_log": web_runner.webdriver_wrapper.get_log,
            "single_quit": web_runner.webdriver_wrapper.quit,
            # web element
            "element_submit": web_runner.webdriver_element.submit,
            "element_clear": web_runner.webdriver_element.clear,
            "element_get_property": web_runner.webdriver_element.get_property,
            "element_get_dom_attribute": web_runner.webdriver_element.get_dom_attribute,
            "element_get_attribute": web_runner.webdriver_element.get_attribute,
            "element_is_selected": web_runner.webdriver_element.is_selected,
            "element_is_enabled": web_runner.webdriver_element.is_enabled,
            "input_to_element": web_runner.webdriver_element.input_to_element,
            "click_element": web_runner.webdriver_element.click_element,
            "element_is_displayed": web_runner.webdriver_element.is_displayed,
            "element_value_of_css_property": web_runner.webdriver_element.value_of_css_property,
            "element_screenshot": web_runner.webdriver_element.screenshot,
            "element_change_web_element": web_runner.webdriver_element.change_web_element,
            "element_check_current_web_element": web_runner.webdriver_element.check_current_web_element,
            "element_get_select": web_runner.webdriver_element.get_select,
            # init test record
            "set_record_enable": test_record_instance.set_record_enable,
            # generate html
            "generate_html": generate_html,
        }

    def _execute_event(self, action: list):
        """
        :param action: execute action
        :return: what event return
        """
        event = self.event_dict.get(action[0])
        if len(action) == 2:
            return event(**action[1])
        elif len(action) == 1:
            return event()
        else:
            raise WebRunnerExecuteException(executor_data_error + " " + str(action))

    def execute_action(self, action_list: [list, dict]) -> dict:
        """
        :param action_list: like this structure
        [

        ]
        for loop and use execute_event function to execute
        :return: recode string, response as list
        """
        if type(action_list) is dict:
            action_list = action_list.get("web_runner", None)
            if action_list is None:
                raise WebRunnerExecuteException(executor_list_error)
        execute_record_dict = dict()
        try:
            if len(action_list) > 0 or type(action_list) is not list:
                pass
            else:
                raise WebRunnerExecuteException(executor_list_error)
        except Exception as error:
            print(repr(error), file=sys.stderr)
        for action in action_list:
            try:
                event_response = self._execute_event(action)
                execute_record = "execute: " + str(action)
                execute_record_dict.update({execute_record: event_response})
            except Exception as error:
                print(repr(error), file=sys.stderr)
                print(action, file=sys.stderr)
        return execute_record_dict

    def execute_files(self, execute_files_list: list) -> list:
        """
        :param execute_files_list: list include execute files path
        :return: every execute detail as list
        """
        execute_detail_list = list()
        for file in execute_files_list:
            execute_detail_list.append(self.execute_action(read_action_json(file)))
        return execute_detail_list


executor = Executor()


def add_command_to_executor(command_dict: dict):
    for command_name, command in command_dict.items():
        if isinstance(command, (types.MethodType, types.FunctionType)):
            executor.event_dict.update({command_name: command})
        else:
            raise WebRunnerAddCommandException(add_command_exception_tag)


def execute_action(action_list: list) -> dict:
    return executor.execute_action(action_list)


def execute_files(execute_files_list: list) -> list:
    return executor.execute_files(execute_files_list)



from typing import Any

from je_web_runner.utils.test_object.test_object_class import TestObject


class TestObjectRecord(object):

    def __init__(self):
        self.test_object_record_dict = dict()

    def clean_record(self) -> None:
        self.test_object_record_dict = dict()

    def save_test_object(self, test_object_name: str, object_type=None, **kwargs) -> None:
        test_object = TestObject(test_object_name, object_type)
        self.test_object_record_dict.update({test_object.test_object_name: test_object})

    def remove_test_object(self, test_object_name: str) -> Any:
        return self.test_object_record_dict.pop(test_object_name, False)


test_object_record = TestObjectRecord()



import datetime
import json
import os
import sys

from je_web_runner.webdriver_download_manager.logger import log
from je_web_runner.webdriver_download_manager.utils import get_date_diff, File, save_file


class DriverCache(object):

    def __init__(self, root_dir=None, valid_range=1):
        self._root_dir = root_dir

        if self._root_dir is None and os.environ.get("WDM_LOCAL", '0') == '1':
            self._root_dir = os.path.join(sys.path[0], ".wdm")
        if self._root_dir is None:
            self._root_dir = os.path.join(os.path.expanduser("~"), ".wdm")
        self._drivers_root = "drivers"
        self._drivers_json_path = os.path.join(self._root_dir, "drivers.json")
        self._date_format = "%d/%m/%Y"
        self._drivers_directory = f"{self._root_dir}{os.sep}{self._drivers_root}"
        self.valid_range = valid_range

    def save_file_to_cache(self, driver, file: File):
        driver_name = driver.get_name()
        os_type = driver.get_os_type()
        driver_version = driver.get_version()
        browser_version = driver.browser_version

        path = os.path.join(self._drivers_directory, driver_name, os_type, driver_version)
        archive = save_file(file, path)
        files = archive.unpack(path)
        binary = self.__get_binary(files, driver_name)
        binary_path = os.path.join(path, binary)
        self.__save_metadata(browser_version, driver_name, os_type, driver_version, binary_path)
        log(f"Driver has been saved in cache [{path}]")
        return binary_path

    def __get_binary(self, files, driver_name):
        if len(files) == 1:
            return files[0]

        for f in files:
            if driver_name in f:
                return f

        raise Exception(f"Can't find binary for {driver_name} among {files}")

    def __save_metadata(self, browser_version, driver_name, os_type, driver_version, binary_path,
                        date=None):
        if date is None:
            date = datetime.date.today()

        metadata = self.get_metadata()

        key = f"{os_type}_{driver_name}_{driver_version}_for_{browser_version}"

        data = {
            key: {
                "timestamp": date.strftime(self._date_format),
                "binary_path": binary_path
            }
        }

        metadata.update(data)
        with open(self._drivers_json_path, 'w+') as outfile:
            json.dump(metadata, outfile, indent=4)

    def find_driver(self, driver):
        """Find driver by '{os_type}_{driver_name}_{driver_version}_{browser_version}'."""
        os_type = driver.get_os_type()
        driver_name = driver.get_name()
        driver_version = driver.get_version()
        browser_version = driver.browser_version

        metadata = self.get_metadata()

        key = f"{os_type}_{driver_name}_{driver_version}_for_{browser_version}"
        if key not in metadata:
            log(f"There is no [{os_type}] {driver_name} for browser {browser_version} in cache")
            return None

        path = os.path.join(self._drivers_directory, driver_name, os_type, driver_version)
        driver_binary_name = 'msedgedriver' if driver_name == 'edgedriver' else driver_name
        driver_binary_name = f'{driver_binary_name}.exe' if 'win' in os_type else driver_name
        binary_path = os.path.join(path, driver_binary_name)
        if not os.path.exists(binary_path):
            return None

        driver_info = metadata[key]

        if not self.__is_valid(driver_info):
            return None

        path = driver_info['binary_path']
        log(f"Driver [{path}] found in cache")
        return path

    def __is_valid(self, driver_info):
        dates_diff = get_date_diff(driver_info['timestamp'],
                                   datetime.date.today(),
                                   self._date_format)
        return dates_diff < self.valid_range

    def get_metadata(self):
        if os.path.exists(self._drivers_json_path):
            with open(self._drivers_json_path, 'r') as outfile:
                return json.load(outfile)
        return {}



import datetime
import os
import platform
import re
import subprocess
import sys

import requests

from je_web_runner.webdriver_download_manager.archive import Archive
from je_web_runner.webdriver_download_manager.logger import log


class File(object):

    def __init__(self, stream):
        self.content = stream.content
        self.__stream = stream
        self.__temp_name = "driver"

    @property
    def filename(self) -> str:
        try:
            filename = re.findall("filename=(.+)", self.__stream.headers["content-disposition"])[0]
        except KeyError:
            filename = f"{self.__temp_name}.zip"
        except IndexError:
            filename = f"{self.__temp_name}.exe"

        if '"' in filename:
            filename = filename.replace('"', "")

        return filename


def save_file(file: File, directory: str):
    os.makedirs(directory, exist_ok=True)

    archive_path = f"{directory}{os.sep}{file.filename}"
    with open(archive_path, "wb") as code:
        code.write(file.content)
    return Archive(archive_path, os_type=os_name())


class OSType(object):
    LINUX = "linux"
    MAC = "mac"
    WIN = "win"


class ChromeType(object):
    GOOGLE = 'google-chrome'
    CHROMIUM = 'chromium'
    BRAVE = "brave-browser"
    MSEDGE = 'edge'


PATTERN = {
    ChromeType.CHROMIUM: r'\d+\.\d+\.\d+',
    ChromeType.GOOGLE: r'\d+\.\d+\.\d+',
    ChromeType.MSEDGE: r'\d+\.\d+\.\d+',
    'brave-browser': r'(\d+)',
    'firefox': r'(\d+.\d+)',
}


def os_name():
    pl = sys.platform
    if pl == "linux" or pl == "linux2":
        return OSType.LINUX
    elif pl == "darwin":
        return OSType.MAC
    elif pl == "win32":
        return OSType.WIN


def os_architecture():
    if platform.machine().endswith('64'):
        return 64
    else:
        return 32


def os_type():
    return os_name() + str(os_architecture())


def validate_response(resp: requests.Response):
    if resp.status_code == 404:
        raise ValueError("There is no such driver by url {}".format(resp.url))
    elif resp.status_code != 200:
        raise ValueError(
            f'response body:\n{resp.text}\n'
            f'request url:\n{resp.request.url}\n'
            f'response headers:\n{dict(resp.headers)}\n'
        )


def download_file(url: str, ssl_verify=True) -> File:
    log(f"Trying to download new driver from {url}")
    response = requests.get(url, stream=True, verify=ssl_verify)
    validate_response(response)
    return File(response)


def get_date_diff(date1, date2, date_format):
    a = datetime.datetime.strptime(date1, date_format)
    b = datetime.datetime.strptime(str(date2.strftime(date_format)), date_format)

    return (b - a).days


def linux_browser_apps_to_cmd(*apps: str) -> str:
    """Create 'browser --version' command from browser app names.

    Result command example:
        chromium --version || chromium-browser --version
    """
    ignore_errors_cmd_part = ' 2>/dev/null' if os.getenv('WDM_LOG_LEVEL') == '0' else ''
    return ' || '.join(f'{i} --version{ignore_errors_cmd_part}' for i in apps)


def windows_browser_apps_to_cmd(*apps: str) -> str:
    """Create analogue of browser --version command for windows."""
    powershell = determine_powershell()

    first_hit_template = """$tmp = {expression}; if ($tmp) {{echo $tmp; Exit;}};"""
    script = "$ErrorActionPreference='silentlycontinue'; " + " ".join(
        first_hit_template.format(expression=e) for e in apps
    )

    return f'{powershell} -NoProfile "{script}"'


def get_browser_version_from_os(browser_type=None):
    """Return installed browser version."""

    pattern = PATTERN[browser_type]

    cmd_mapping = {
        ChromeType.GOOGLE: {
            OSType.LINUX: linux_browser_apps_to_cmd('google-chrome', 'google-chrome-stable', 'google-chrome-beta',
                                                    'google-chrome-dev'),
            OSType.MAC: r'/Applications/Google\ Chrome.app/Contents/MacOS/Google\ Chrome --version',
            OSType.WIN: windows_browser_apps_to_cmd(
                r'(Get-Item -Path "$env:PROGRAMFILES\Google\Chrome\Application\chrome.exe").VersionInfo.FileVersion',
                r'(Get-Item -Path "$env:PROGRAMFILES (x86)\Google\Chrome\Application\chrome.exe").VersionInfo.FileVersion',
                r'(Get-Item -Path "$env:LOCALAPPDATA\Google\Chrome\Application\chrome.exe").VersionInfo.FileVersion',
                r'(Get-ItemProperty -Path Registry::"HKCU\SOFTWARE\Google\Chrome\BLBeacon").version',
                r'(Get-ItemProperty -Path Registry::"HKLM\SOFTWARE\Wow6432Node\Microsoft\Windows\CurrentVersion\Uninstall\Google Chrome").version'
            ),
        },
        ChromeType.CHROMIUM: {
            OSType.LINUX: linux_browser_apps_to_cmd('chromium', 'chromium-browser'),
            OSType.MAC: r'/Applications/Chromium.app/Contents/MacOS/Chromium --version',
            OSType.WIN: windows_browser_apps_to_cmd(
                r'(Get-Item -Path "$env:PROGRAMFILES\Chromium\Application\chrome.exe").VersionInfo.FileVersion',
                r'(Get-Item -Path "$env:PROGRAMFILES (x86)\Chromium\Application\chrome.exe").VersionInfo.FileVersion',
                r'(Get-Item -Path "$env:LOCALAPPDATA\Chromium\Application\chrome.exe").VersionInfo.FileVersion',
                r'(Get-ItemProperty -Path Registry::"HKCU\SOFTWARE\Chromium\BLBeacon").version',
                r'(Get-ItemProperty -Path Registry::"HKLM\SOFTWARE\Wow6432Node\Microsoft\Windows\CurrentVersion\Uninstall\Chromium").version'
            ),
        },
        ChromeType.BRAVE: {
            OSType.LINUX: linux_browser_apps_to_cmd('brave-browser', 'brave-browser-beta', 'brave-browser-nightly'),
            OSType.MAC: r'/Applications/Brave\ Browser.app/Contents/MacOS/Brave\ Browser --version',
            OSType.WIN: windows_browser_apps_to_cmd(
                r'(Get-Item -Path "$env:PROGRAMFILES\BraveSoftware\Brave-Browser\Application\brave.exe").VersionInfo.FileVersion',
                r'(Get-Item -Path "$env:PROGRAMFILES (x86)\BraveSoftware\Brave-Browser\Application\brave.exe").VersionInfo.FileVersion',
                r'(Get-Item -Path "$env:LOCALAPPDATA\BraveSoftware\Brave-Browser\Application\brave.exe").VersionInfo.FileVersion',
                r'(Get-ItemProperty -Path Registry::"HKCU\SOFTWARE\BraveSoftware\Brave-Browser\BLBeacon").version',
                r'(Get-ItemProperty -Path Registry::"HKLM\SOFTWARE\Wow6432Node\Microsoft\Windows\CurrentVersion\Uninstall\BraveSoftware Brave-Browser").version'
            ),
        },
        ChromeType.MSEDGE: {
            OSType.LINUX: linux_browser_apps_to_cmd('microsoft-edge', 'microsoft-edge-stable', 'microsoft-edge-beta',
                                                    'microsoft-edge-dev'),
            OSType.MAC: r'/Applications/Microsoft\ Edge.app/Contents/MacOS/Microsoft\ Edge --version',
            OSType.WIN: windows_browser_apps_to_cmd(
                # stable edge
                r'(Get-Item -Path "$env:PROGRAMFILES\Microsoft\Edge\Application\msedge.exe").VersionInfo.FileVersion',
                r'(Get-Item -Path "$env:PROGRAMFILES (x86)\Microsoft\Edge\Application\msedge.exe").VersionInfo.FileVersion',
                r'(Get-ItemProperty -Path Registry::"HKCU\SOFTWARE\Microsoft\Edge\BLBeacon").version',
                r'(Get-ItemProperty -Path Registry::"HKLM\SOFTWARE\Microsoft\EdgeUpdate\Clients\{56EB18F8-8008-4CBD-B6D2-8C97FE7E9062}").pv',
                # beta edge
                r'(Get-Item -Path "$env:LOCALAPPDATA\Microsoft\Edge Beta\Application\msedge.exe").VersionInfo.FileVersion',
                r'(Get-Item -Path "$env:PROGRAMFILES\Microsoft\Edge Beta\Application\msedge.exe").VersionInfo.FileVersion',
                r'(Get-Item -Path "$env:PROGRAMFILES (x86)\Microsoft\Edge Beta\Application\msedge.exe").VersionInfo.FileVersion',
                r'(Get-ItemProperty -Path Registry::"HKCU\SOFTWARE\Microsoft\Edge Beta\BLBeacon").version',
                # dev edge
                r'(Get-Item -Path "$env:LOCALAPPDATA\Microsoft\Edge Dev\Application\msedge.exe").VersionInfo.FileVersion',
                r'(Get-Item -Path "$env:PROGRAMFILES\Microsoft\Edge Dev\Application\msedge.exe").VersionInfo.FileVersion',
                r'(Get-Item -Path "$env:PROGRAMFILES (x86)\Microsoft\Edge Dev\Application\msedge.exe").VersionInfo.FileVersion',
                r'(Get-ItemProperty -Path Registry::"HKCU\SOFTWARE\Microsoft\Edge Dev\BLBeacon").version',
                # canary edge
                r'(Get-Item -Path "$env:LOCALAPPDATA\Microsoft\Edge SxS\Application\msedge.exe").VersionInfo.FileVersion',
                r'(Get-ItemProperty -Path Registry::"HKCU\SOFTWARE\Microsoft\Edge SxS\BLBeacon").version',
                # highest edge
                r"(Get-Item (Get-ItemProperty 'HKLM:\SOFTWARE\Microsoft\Windows\CurrentVersion\App Paths\msedge.exe').'(Default)').VersionInfo.ProductVersion",
                r"[System.Diagnostics.FileVersionInfo]::GetVersionInfo((Get-ItemProperty 'HKLM:\SOFTWARE\Microsoft\Windows\CurrentVersion\App Paths\msedge.exe').'(Default)').ProductVersion",
                r'Get-AppxPackage -Name *MicrosoftEdge.* | Foreach Version',
                r'(Get-ItemProperty -Path Registry::"HKLM\SOFTWARE\Wow6432Node\Microsoft\Windows\CurrentVersion\Uninstall\Microsoft Edge").version'
            ),
        },
        'firefox': {
            OSType.LINUX: linux_browser_apps_to_cmd('firefox'),
            OSType.MAC: r'/Applications/Firefox.app/Contents/MacOS/firefox --version',
            OSType.WIN: windows_browser_apps_to_cmd(
                r'(Get-Item -Path "$env:PROGRAMFILES\Mozilla Firefox\firefox.exe").VersionInfo.FileVersion',
                r'(Get-Item -Path "$env:PROGRAMFILES (x86)\Mozilla Firefox\firefox.exe").VersionInfo.FileVersion',
                r"(Get-Item (Get-ItemProperty 'HKLM:\SOFTWARE\Microsoft\Windows\CurrentVersion\App Paths\firefox.exe').'(Default)').VersionInfo.ProductVersion",
                r'(Get-ItemProperty -Path Registry::"HKLM\SOFTWARE\Mozilla\Mozilla Firefox").CurrentVersion'
            ),
        },
    }[browser_type][os_name()]

    version = read_version_from_cmd(cmd_mapping, pattern)

    if not version:
        log(
            f'Could not get version for {browser_type}.'
            f'Is {browser_type} installed?'
        )
    else:
        log(f"Current {browser_type} version is {version}")

    return version


def read_version_from_cmd(cmd, pattern):
    with subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            stdin=subprocess.DEVNULL,
            shell=True,
    ) as stream:
        stdout = stream.communicate()[0].decode()
        version = re.search(pattern, stdout)
        version = version.group(0) if version else None
    return version


def determine_powershell():
    """Returns "True" if runs in Powershell and "False" if another console."""
    cmd = '(dir 2>&1 *`|echo CMD);&<# rem #>echo powershell'
    with subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            stdin=subprocess.DEVNULL,
            shell=True,
    ) as stream:
        stdout = stream.communicate()[0].decode()
    return "" if stdout == 'powershell' else "powershell"



import logging
import os

loggers = {}


def _init_logger(level=logging.INFO, name="WDM", first_line=False, formatter='[%(name)s] - %(message)s'):
    """Initialize the logger."""
    log_level = os.getenv('WDM_LOG_LEVEL')
    if log_level:
        level = int(log_level)
    if not loggers.get(name):
        _logger = logging.getLogger(name)

        handler = logging.StreamHandler()
        formatter = logging.Formatter(formatter)
        handler.setFormatter(formatter)
        _logger.addHandler(handler)
        _logger.setLevel(level)
        loggers[name] = _logger


def log(text, level=logging.INFO, name="WDM", first_line=False, formatter='[%(name)s] - %(message)s'):
    """Emitting the log message."""
    if os.getenv('WDM_LOG', None) != '0':
        _init_logger(level, name, first_line, formatter)
        loggers.get(name).info(text)



from sys import stderr
from typing import List

from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support.ui import Select

from je_web_runner.utils.assert_value.result_check import check_web_element
from je_web_runner.utils.test_record.test_record_class import record_action_to_list


class WebElementWrapper(object):

    def __init__(self):
        self.current_web_element: [WebElement] = None
        self.current_web_element_list: [List[WebElement]] = None

    def submit(self) -> None:
        """
        current web element submit
        :return: None
        """
        try:
            self.current_web_element.submit()
            record_action_to_list("Web element submit", None, None)
        except Exception as error:
            print(repr(error), file=stderr)
            record_action_to_list("Web element submit", None, error)

    def clear(self) -> None:
        """
        current web element clear
        :return: None
        """
        try:
            self.current_web_element.clear()
            record_action_to_list("Web element clear", None, None)
        except Exception as error:
            print(repr(error), file=stderr)
            record_action_to_list("Web element clear", None, error)

    def get_property(self, name) -> str:
        """
        :param name: name of property
        :return: property value as str
        """
        param = locals()
        try:
            record_action_to_list("Web element get_property", param, None)
            return self.current_web_element.get_property(name)
        except Exception as error:
            print(repr(error), file=stderr)
            record_action_to_list("Web element get_property", param, error)

    def get_dom_attribute(self, name) -> str:
        """
        :param name: name of dom
        :return: dom attribute value as str
        """
        param = locals()
        try:
            record_action_to_list("Web element get_dom_attribute", param, None)
            return self.current_web_element.get_dom_attribute(name)
        except Exception as error:
            print(repr(error), file=stderr)
            record_action_to_list("Web element get_dom_attribute", param, error)

    def get_attribute(self, name) -> str:
        """
        :param name: name of web element
        :return:web element attribute value as str
        """
        param = locals()
        try:
            record_action_to_list("Web element get_attribute", param, None)
            return self.current_web_element.get_attribute(name)
        except Exception as error:
            print(repr(error), file=stderr)
            record_action_to_list("Web element get_attribute", param, error)

    def is_selected(self) -> bool:
        """
        check current web element is selected or not
        :return: True or False
        """
        try:
            record_action_to_list("Web element is_selected", None, None)
            return self.current_web_element.is_selected()
        except Exception as error:
            print(repr(error), file=stderr)
            record_action_to_list("Web element is_selected", None, error)

    def is_enabled(self) -> bool:
        """
        check current web element is enable or not
        :return: True or False
        """
        try:
            record_action_to_list("Web element is_enabled", None, None)
            return self.current_web_element.is_enabled()
        except Exception as error:
            print(repr(error), file=stderr)
            record_action_to_list("Web element is_enabled", None, error)

    def input_to_element(self, input_value) -> None:
        """
        input value to current web element
        :param input_value: what value we want to input to current web element
        :return: None
        """
        param = locals()
        try:
            self.current_web_element.send_keys(input_value)
            record_action_to_list("Web element input_to_element", param, None)
        except Exception as error:
            print(repr(error), file=stderr)
            record_action_to_list("Web element input_to_element", param, error)

    def click_element(self) -> None:
        """
        click current web element
        :return: None
        """
        try:
            self.current_web_element.click()
            record_action_to_list("Web element click_element", None, None)
        except Exception as error:
            print(repr(error), file=stderr)
            record_action_to_list("Web element click_element", None, error)

    def is_displayed(self) -> bool:
        """
        check current web element is displayed or not
        :return: True or False
        """
        try:
            record_action_to_list("Web element is_displayed", None, None)
            return self.current_web_element.is_displayed()
        except Exception as error:
            print(repr(error), file=stderr)
            record_action_to_list("Web element is_displayed", None, error)

    def value_of_css_property(self, property_name) -> str:
        """
        :param property_name: name of property
        :return: css property value as str
        """
        param = locals()
        try:
            record_action_to_list("Web element value_of_css_property", param, None)
            return self.current_web_element.value_of_css_property(property_name)
        except Exception as error:
            print(repr(error), file=stderr)
            record_action_to_list("Web element value_of_css_property", param, error)

    def screenshot(self, filename) -> bool:
        """
        :param filename: full file name not need .png extension
        :return: Save True or not
        """
        param = locals()
        try:
            record_action_to_list("Web element screenshot", param, None)
            return self.current_web_element.screenshot(filename + ".png")
        except Exception as error:
            print(repr(error), file=stderr)
            record_action_to_list("Web element screenshot", param, error)

    # Web element wrapper add function
    def change_web_element(self, element_index: int) -> None:
        """
        :param element_index: change to web element index
        :return: web element list [element_index]
        """
        param = locals()
        try:
            self.current_web_element = self.current_web_element_list[element_index]
            record_action_to_list("Web element change_web_element", param, None)
        except Exception as error:
            print(repr(error), file=stderr)
            record_action_to_list("Web element change_web_element", param, error)

    def check_current_web_element(self, check_dict: dict) -> None:
        """
        :param check_dict: check web element dict {name: should be value}
        :return: None
        """
        param = locals()
        try:
            check_web_element(self.current_web_element, check_dict)
            record_action_to_list("Web element check_current_web_element", param, None)
        except Exception as error:
            print(repr(error), file=stderr)
            record_action_to_list("Web element check_current_web_element", param, error)

    def get_select(self) -> Select:
        """
        get Select(current web element)
        :return: Select(current web element)
        """
        try:
            record_action_to_list("Web element get_select", None, None)
            return Select(self.current_web_element)
        except Exception as error:
            print(repr(error), file=stderr)
            record_action_to_list("Web element get_select", None, error)


web_element_wrapper = WebElementWrapper()



from sys import stderr
from typing import Union

from selenium import webdriver
from selenium.webdriver.chrome import options
from selenium.webdriver.edge import options
from selenium.webdriver.firefox import options
from selenium.webdriver.ie import options

from je_web_runner.utils.exception.exception_tag import selenium_wrapper_set_argument_error
from je_web_runner.utils.exception.exception_tag import selenium_wrapper_set_options_error
from je_web_runner.utils.exception.exception_tag import selenium_wrapper_web_driver_not_found_error
from je_web_runner.utils.exception.exceptions import WebRunnerArgumentWrongTypeException
from je_web_runner.utils.exception.exceptions import WebRunnerOptionsWrongTypeException
from je_web_runner.utils.exception.exceptions import WebRunnerWebDriverNotFoundException
from je_web_runner.utils.test_record.test_record_class import record_action_to_list

selenium_options_dict = {
    "chrome": webdriver.chrome.options.Options,
    "chromium": webdriver.chrome.options.Options,
    "firefox": webdriver.firefox.options.Options,
    "edge": webdriver.edge.options.Options,
    "ie": webdriver.ie.options.Options,
}


def set_webdriver_options_argument(webdriver_name: str, argument_iterable: [list, set]) -> \
        Union[
            webdriver.chrome.options.Options, webdriver.chrome.options.Options,
            webdriver.edge.options.Options, webdriver.ie.options.Options
        ]:
    param = locals()
    try:
        webdriver_options = selenium_options_dict.get(webdriver_name)()
        for i in range(len(argument_iterable)):
            if type(argument_iterable[i]) != str:
                raise WebRunnerArgumentWrongTypeException(selenium_wrapper_set_argument_error)
            webdriver_options.add_argument(argument_iterable[i])
        record_action_to_list("webdriver with options set_webdriver_options_argument", param, None)
        return webdriver_options
    except Exception as error:
        print(repr(error), file=stderr)
        record_action_to_list("webdriver with options set_webdriver_options_argument", param, error)


def set_webdriver_options_capability_wrapper(webdriver_name: str, key_and_vale_dict: dict) -> \
        Union[
            webdriver.chrome.options.Options, webdriver.chrome.options.Options,
            webdriver.edge.options.Options, webdriver.ie.options.Options
        ]:
    param = locals()
    try:
        webdriver_options = selenium_options_dict.get(webdriver_name)()
        if webdriver_options is None:
            raise WebRunnerWebDriverNotFoundException(selenium_wrapper_web_driver_not_found_error)
        if type(key_and_vale_dict) is not dict:
            raise WebRunnerOptionsWrongTypeException(selenium_wrapper_set_options_error)
        for key, value in key_and_vale_dict.items():
            webdriver_options.set_capability(key, value)
        record_action_to_list("webdriver with options set_webdriver_options_capability_wrapper", param, None)
        return webdriver_options
    except Exception as error:
        print(repr(error), file=stderr)
        record_action_to_list("webdriver with options set_webdriver_options_capability_wrapper", param, error)



import os
import platform

import requests

from je_web_runner.webdriver_download_manager.logger import log
from je_web_runner.webdriver_download_manager.utils import (
    validate_response,
    get_browser_version_from_os,
    ChromeType,
    OSType,
)


class Driver(object):
    def __init__(self, name,
                 version,
                 os_type,
                 url,
                 latest_release_url):
        self._name = name
        self._url = url
        self._version = version
        self._os_type = os_type
        self._latest_release_url = latest_release_url
        self.ssl_verify = False if os.getenv('WDM_SSL_VERIFY') == '0' else True

    def get_name(self):
        return self._name

    def get_os_type(self):
        return self._os_type

    def get_url(self):
        return f"{self._url}/{self.get_version()}/{self.get_name()}_{self.get_os_type()}.zip"

    def get_version(self):
        self._version = (
            self.get_latest_release_version()
            if self._version == "latest"
            else self._version
        )
        return self._version

    def get_latest_release_version(self):
        # type: () -> str
        raise NotImplementedError("Please implement this method")


class ChromeDriver(Driver):
    def __init__(self, name, version, os_type, url, latest_release_url,
                 chrome_type=ChromeType.GOOGLE):
        super(ChromeDriver, self).__init__(name, version, os_type, url,
                                           latest_release_url)
        self.chrome_type = chrome_type
        self.browser_version = ""

    def get_os_type(self):
        if "win" in super().get_os_type():
            return "win32"
        mac = f'{super().get_os_type()}{"_m1" if "mac" in super().get_os_type() and not platform.processor() == "i386" else ""}'
        return mac if 'mac' in super().get_os_type() else super().get_os_type()

    def get_latest_release_version(self):
        self.browser_version = get_browser_version_from_os(self.chrome_type)
        log(f"Get LATEST {self._name} version for {self.browser_version} {self.chrome_type}")
        latest_release_url = (
            f"{self._latest_release_url}_{self.browser_version}"
            if self.browser_version
            else self._latest_release_url
        )
        resp = requests.get(
            url=latest_release_url,
            verify=self.ssl_verify
        )
        validate_response(resp)
        self._version = resp.text.rstrip()
        return self._version


class GeckoDriver(Driver):
    def __init__(
            self,
            name,
            version,
            os_type,
            url,
            latest_release_url,
            mozila_release_tag,
    ):
        super(GeckoDriver, self).__init__(
            name,
            version,
            os_type,
            url,
            latest_release_url,
        )
        self._mozila_release_tag = mozila_release_tag
        self.browser_version = ""
        self._os_token = os.getenv("GH_TOKEN", None)
        self.auth_header = (
            {'Authorization': f'token {self._os_token}'}
            if self._os_token
            else None
        )
        if self._os_token:
            log("GH_TOKEN will be used to perform requests", first_line=True)

    def get_latest_release_version(self) -> str:
        self.browser_version = get_browser_version_from_os("firefox")
        log(f"Get LATEST {self._name} version for {self.browser_version} firefox")
        resp = requests.get(
            url=self.latest_release_url,
            headers=self.auth_header,
            verify=self.ssl_verify,
        )
        validate_response(resp)
        self._version = resp.json()["tag_name"]
        return self._version

    def get_url(self):
        """Like https://github.com/mozilla/geckodriver/releases/download/v0.11.1/geckodriver-v0.11.1-linux64.tar.gz"""
        log(f"Getting latest mozilla release info for {self.get_version()}")
        resp = requests.get(
            url=self.tagged_release_url(self.get_version()),
            headers=self.auth_header,
            verify=self.ssl_verify,
        )
        validate_response(resp)
        assets = resp.json()["assets"]

        name = f"{self.get_name()}-{self.get_version()}-{self.get_os_type()}."
        output_dict = [
            asset for asset in assets if asset['name'].startswith(name)
        ]
        return output_dict[0]['browser_download_url']

    def get_os_type(self):
        mac = f'macos{"-aarch64" if platform.processor() != "i386" else ""}'
        return mac if 'mac' in super().get_os_type() else super().get_os_type()

    @property
    def latest_release_url(self):
        return self._latest_release_url

    def tagged_release_url(self, version):
        return self._mozila_release_tag.format(version)


class IEDriver(Driver):
    def __init__(
            self,
            name,
            version,
            os_type,
            url,
            latest_release_url,
            ie_release_tag,
    ):
        super(IEDriver, self).__init__(
            name,
            version,
            os_type,
            url,
            latest_release_url,
        )
        self.os_type = "x64" if os_type == "win64" else "Win32"
        self._ie_release_tag = ie_release_tag
        # todo: for 'browser_version' implement installed IE version detection
        #       like chrome or firefox
        self.browser_version = ""
        self._os_token = os.getenv("GH_TOKEN", None)
        self.auth_header = (
            {'Authorization': f'token {self._os_token}'}
            if self._os_token
            else None
        )
        if self._os_token:
            log("GH_TOKEN will be used to perform requests", first_line=True)

    def get_latest_release_version(self) -> str:
        log(f"Get LATEST driver version for {self.browser_version}")
        resp = requests.get(
            url=self.latest_release_url,
            headers=self.auth_header,
            verify=self.ssl_verify,
        )
        validate_response(resp)
        releases = resp.json()
        release = next(
            release
            for release in releases
            for asset in release['assets']
            if asset['name'].startswith(self.get_name())
        )
        self._version = release['tag_name'].replace('selenium-', '')
        return self._version

    def get_url(self):
        """Like https://github.com/seleniumhq/selenium/releases/download/3.141.59/IEDriverServer_Win32_3.141.59.zip"""
        log(f"Getting latest ie release info for {self.get_version()}")
        resp = requests.get(
            url=self.tagged_release_url(self.get_version()),
            headers=self.auth_header,
            verify=self.ssl_verify,
        )
        validate_response(resp)
        assets = resp.json()["assets"]

        name = f"{self.get_name()}_{self.os_type}_{self.get_version()}" + "."
        output_dict = [
            asset for asset in assets if asset['name'].startswith(name)
        ]
        return output_dict[0]['browser_download_url']

    @property
    def latest_release_url(self):
        return self._latest_release_url

    def tagged_release_url(self, version):
        version = self.__get_divided_version(version)
        return self._ie_release_tag.format(version)

    def __get_divided_version(self, version):
        divided_version = version.split('.')
        if len(divided_version) == 2:
            return f'{version}.0'
        elif len(divided_version) == 3:
            return version
        else:
            raise ValueError(
                "Version must consist of major, minor and/or patch, "
                "but given was: {version}".format(version=self.get_version())
            )


class EdgeChromiumDriver(Driver):
    def __init__(
            self,
            name,
            version,
            os_type,
            url,
            latest_release_url,
    ):
        super(EdgeChromiumDriver, self).__init__(
            name,
            version,
            os_type,
            url,
            latest_release_url,
        )
        self.browser_version = ""

    def get_stable_release_version(self):
        """Stable driver version when browser version was not determined."""
        stable = self._latest_release_url.replace('LATEST_RELEASE', 'LATEST_STABLE')
        resp = requests.get(stable, verify=self.ssl_verify)
        validate_response(resp)
        return resp.text.rstrip()

    def get_latest_release_version(self) -> str:
        browser_version = get_browser_version_from_os(ChromeType.MSEDGE)
        self.browser_version = (
            browser_version
            if browser_version
            else self.get_stable_release_version()
        )
        log(f"Get LATEST {self._name} version for {self.browser_version} Edge")
        major_edge_version = self.browser_version.split(".")[0]
        latest_release_url = {
            OSType.WIN in self.get_os_type(): f'{self._latest_release_url}_{major_edge_version}_WINDOWS',
            OSType.MAC in self.get_os_type(): f'{self._latest_release_url}_{major_edge_version}_MACOS',
            OSType.LINUX in self.get_os_type(): f'{self._latest_release_url}_{major_edge_version}_LINUX',
        }[True]
        resp = requests.get(latest_release_url, verify=self.ssl_verify)
        validate_response(resp)
        self._version = resp.text.rstrip()
        return self._version



import logging
import os

from je_web_runner.webdriver_download_manager import utils
from je_web_runner.webdriver_download_manager.driver import EdgeChromiumDriver
from je_web_runner.webdriver_download_manager.driver import IEDriver
from je_web_runner.webdriver_download_manager.manager import DriverManager


class IEDriverManager(DriverManager):
    def __init__(
            self,
            version="latest",
            os_type=utils.os_type(),
            path=None,
            name="IEDriverServer",
            url="https://github.com/seleniumhq/selenium/releases/download",
            latest_release_url="https://api.github.com/repos/seleniumhq/selenium/releases",
            ie_release_tag="https://api.github.com/repos/seleniumhq/selenium/releases/tags/selenium-{0}",
            log_level=logging.INFO,
            print_first_line=True,
            cache_valid_range=1,
    ):
        super().__init__(path, log_level, print_first_line, cache_valid_range)
        self.driver = IEDriver(
            version=version,
            os_type=os_type,
            name=name,
            url=url,
            latest_release_url=latest_release_url,
            ie_release_tag=ie_release_tag,
        )

    def install(self):
        return self._get_driver_path(self.driver)


class EdgeChromiumDriverManager(DriverManager):
    def __init__(
            self,
            version="latest",
            os_type=utils.os_type(),
            path=None,
            name="edgedriver",
            url="https://msedgedriver.azureedge.net",
            latest_release_url="https://msedgedriver.azureedge.net/LATEST_RELEASE",
            log_level=logging.INFO,
            print_first_line=None,
            cache_valid_range=1,
    ):
        super().__init__(path, log_level, print_first_line, cache_valid_range)
        self.driver = EdgeChromiumDriver(
            version=version,
            os_type=os_type,
            name=name,
            url=url,
            latest_release_url=latest_release_url,
        )

    def install(self):
        driver_path = self._get_driver_path(self.driver)

        os.chmod(driver_path, 0o755)
        return driver_path



import os

from je_web_runner.webdriver_download_manager.driver_cache import DriverCache
from je_web_runner.webdriver_download_manager.logger import log
from je_web_runner.webdriver_download_manager.utils import download_file


class DriverManager(object):
    def __init__(self, root_dir=None, log_level=None, print_first_line=None, cache_valid_range=1):
        self.driver_cache = DriverCache(root_dir, cache_valid_range)
        if os.environ.get('WDM_PRINT_FIRST_LINE', str(print_first_line)) == 'True':
            log("\n", formatter='%(message)s', level=log_level)

    def install(self):
        raise NotImplementedError("Please Implement this method")

    def _get_driver_path(self, driver):
        binary_path = self.driver_cache.find_driver(driver)
        if binary_path:
            return binary_path

        file = download_file(driver.get_url(), driver.ssl_verify)
        binary_path = self.driver_cache.save_file_to_cache(driver, file)
        return binary_path



import logging
import os

from je_web_runner.webdriver_download_manager import utils
from je_web_runner.webdriver_download_manager.driver import ChromeDriver
from je_web_runner.webdriver_download_manager.manager import DriverManager
from je_web_runner.webdriver_download_manager.utils import ChromeType


class ChromeDriverManager(DriverManager):
    def __init__(self, version="latest",
                 os_type=utils.os_type(),
                 path=None,
                 name="chromedriver",
                 url="https://chromedriver.storage.googleapis.com",
                 latest_release_url="https://chromedriver.storage.googleapis.com/LATEST_RELEASE",
                 chrome_type=ChromeType.GOOGLE,
                 log_level=logging.INFO,
                 print_first_line=True,
                 cache_valid_range=1):
        super().__init__(path, log_level=log_level, print_first_line=print_first_line,
                         cache_valid_range=cache_valid_range)

        self.driver = ChromeDriver(name=name,
                                   version=version,
                                   os_type=os_type,
                                   url=url,
                                   latest_release_url=latest_release_url,
                                   chrome_type=chrome_type)

    def install(self):
        driver_path = self._get_driver_path(self.driver)

        os.chmod(driver_path, 0o755)
        return driver_path



import logging
import os

from je_web_runner.webdriver_download_manager import utils
from je_web_runner.webdriver_download_manager.driver import GeckoDriver
from je_web_runner.webdriver_download_manager.manager import DriverManager


class GeckoDriverManager(DriverManager):
    def __init__(self, version="latest",
                 os_type=utils.os_type(),
                 path=None,
                 name="geckodriver",
                 url="https://github.com/mozilla/geckodriver/releases/download",
                 latest_release_url="https://api.github.com/repos/mozilla/geckodriver/releases/latest",
                 mozila_release_tag="https://api.github.com/repos/mozilla/geckodriver/releases/tags/{0}",
                 log_level=logging.INFO,
                 print_first_line=True,
                 cache_valid_range=1):
        super(GeckoDriverManager, self).__init__(path, log_level, print_first_line, cache_valid_range)

        self.driver = GeckoDriver(version=version,
                                  os_type=os_type,
                                  name=name,
                                  url=url,
                                  latest_release_url=latest_release_url,
                                  mozila_release_tag=mozila_release_tag)

    def install(self):
        driver_path = self._get_driver_path(self.driver)

        os.chmod(driver_path, 0o755)
        return driver_path



from typing import Any, Union

from selenium.webdriver import DesiredCapabilities

from je_web_runner.utils.exception.exception_tag import selenium_wrapper_web_driver_not_found_error
from je_web_runner.utils.exception.exceptions import WebRunnerException

desired_capabilities_dict = {
    "firefox": DesiredCapabilities.FIREFOX,
    "chrome": DesiredCapabilities.CHROME,
    "edge": DesiredCapabilities.EDGE,
    "safari": DesiredCapabilities.SAFARI,
}


def get_desired_capabilities_keys() -> Union[str, Any]:
    return desired_capabilities_dict.keys()


def get_desired_capabilities(webdriver_name: str) -> \
        [
            DesiredCapabilities.FIREFOX.copy(),
            DesiredCapabilities.CHROME.copy(),
            DesiredCapabilities.EDGE.copy(),
            DesiredCapabilities.SAFARI.copy(),
        ]:
    desired_capabilities = desired_capabilities_dict.get(webdriver_name)
    if desired_capabilities is None:
        raise WebRunnerException(selenium_wrapper_web_driver_not_found_error)
    return desired_capabilities.copy()
